import { SafeString } from 'handlebars';
import { isAbsolute, resolve } from 'path';
import { URL } from 'url';
import { brotliDecompressSync, inflateSync, unzipSync } from 'zlib';
/**
 * Transform http headers objects to Mockoon's Header key value object
 *
 * @param object
 */
const TransformHeaders = (headers) => Object.keys(headers).reduce((newHeaders, key) => {
    const headerValue = headers[key];
    let value = '';
    if (headerValue !== undefined) {
        if (Array.isArray(headerValue)) {
            value = headerValue.join(',');
        }
        else {
            value = headerValue.toString();
        }
    }
    newHeaders.push({ key, value });
    return newHeaders;
}, []);
/**
 * Sort by ascending order
 *
 * @param a
 * @param b
 */
const AscSort = (a, b) => {
    if (a.key < b.key) {
        return -1;
    }
    else {
        return 1;
    }
};
/**
 * Check if an Object or Array is empty
 *
 * @param obj
 */
export const IsEmpty = (obj) => [Object, Array].includes((obj || {}).constructor) &&
    !Object.entries(obj || {}).length;
/**
 * Return a random integer
 *
 * @param a
 * @param b
 */
export const RandomInt = (a = 1, b = 0) => {
    const lower = Math.ceil(Math.min(a, b));
    const upper = Math.floor(Math.max(a, b));
    return Math.floor(lower + Math.random() * (upper - lower + 1));
};
/**
 * Decompress body based on content-encoding
 *
 * @param response
 */
export const DecompressBody = (response) => {
    if (!response.body) {
        return response.body;
    }
    const contentEncoding = response.getHeader('content-encoding');
    let body = response.body;
    switch (contentEncoding) {
        case 'gzip':
            body = unzipSync(body);
            break;
        case 'br':
            body = brotliDecompressSync(body);
            break;
        case 'deflate':
            body = inflateSync(body);
            break;
        default:
            break;
    }
    return body.toString('utf-8');
};
/**
 * Create a Transaction object from express req/res.
 * To be used after the response closes
 *
 * @param request
 * @param response
 */
export function CreateTransaction(request, response) {
    const requestUrl = new URL(request.originalUrl, 'http://localhost/');
    return {
        request: {
            method: request.method.toLowerCase(),
            urlPath: requestUrl.pathname,
            route: request.route ? request.route.path : null,
            params: request.params
                ? Object.keys(request.params).map((paramName) => ({
                    name: paramName,
                    value: request.params[paramName]
                }))
                : [],
            query: requestUrl ? decodeURI(requestUrl.search.slice(1)) : null,
            queryParams: request.query,
            body: request.stringBody,
            headers: TransformHeaders(request.headers).sort(AscSort)
        },
        response: {
            statusCode: response.statusCode,
            headers: TransformHeaders(response.getHeaders()).sort(AscSort),
            body: DecompressBody(response)
        },
        routeResponseUUID: response.routeResponseUUID,
        routeUUID: response.routeUUID,
        proxied: request.proxied || false
    };
}
/**
 * Convert a string to base64
 *
 * @param text
 */
export const ToBase64 = (text) => {
    if (typeof btoa === 'function') {
        return btoa(text);
    }
    if (typeof Buffer === 'function') {
        return Buffer.from(text).toString('base64');
    }
    return text;
};
/**
 * Convert a SafeString to a string if needed.
 *
 * @param text
 * @returns
 */
export const fromSafeString = (text) => text instanceof SafeString ? text.toString() : text;
/**
 * Parse a number from a SafeString if needed.
 *
 * @param text
 * @returns
 */
export const numberFromSafeString = (text) => {
    const parsedText = text instanceof SafeString ? text.toString() : text;
    return parseInt(parsedText, 10);
};
/**
 * Resolve a file path relatively to the current environment folder if provided
 */
export const resolvePathFromEnvironment = (filePath, environmentDirectory) => {
    if (environmentDirectory && !isAbsolute(filePath)) {
        return resolve(environmentDirectory, filePath);
    }
    return filePath;
};
//# sourceMappingURL=utils.js.map